# Copyright: (c) 2023, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

from libc.stdlib cimport calloc, free, malloc

import dataclasses
import enum


cdef extern from "python_sspi.h":
    unsigned int _SECBUFFER_VERSION "SECBUFFER_VERSION"

    unsigned int _SECBUFFER_EMPTY "SECBUFFER_EMPTY"
    unsigned int _SECBUFFER_DATA "SECBUFFER_DATA"
    unsigned int _SECBUFFER_TOKEN "SECBUFFER_TOKEN"
    unsigned int _SECBUFFER_PKG_PARAMS "SECBUFFER_PKG_PARAMS"
    unsigned int _SECBUFFER_MISSING "SECBUFFER_MISSING"
    unsigned int _SECBUFFER_EXTRA "SECBUFFER_EXTRA"
    unsigned int _SECBUFFER_STREAM_TRAILER "SECBUFFER_STREAM_TRAILER"
    unsigned int _SECBUFFER_STREAM_HEADER "SECBUFFER_STREAM_HEADER"
    unsigned int _SECBUFFER_NEGOTIATION_INFO "SECBUFFER_NEGOTIATION_INFO"
    unsigned int _SECBUFFER_PADDING "SECBUFFER_PADDING"
    unsigned int _SECBUFFER_STREAM "SECBUFFER_STREAM"
    unsigned int _SECBUFFER_MECHLIST "SECBUFFER_MECHLIST"
    unsigned int _SECBUFFER_MECHLIST_SIGNATURE "SECBUFFER_MECHLIST_SIGNATURE"
    unsigned int _SECBUFFER_TARGET "SECBUFFER_TARGET"
    unsigned int _SECBUFFER_CHANNEL_BINDINGS "SECBUFFER_CHANNEL_BINDINGS"
    unsigned int _SECBUFFER_CHANGE_PASS_RESPONSE "SECBUFFER_CHANGE_PASS_RESPONSE"
    unsigned int _SECBUFFER_TARGET_HOST "SECBUFFER_TARGET_HOST"
    unsigned int _SECBUFFER_ALERT "SECBUFFER_ALERT"
    unsigned int _SECBUFFER_APPLICATION_PROTOCOLS "SECBUFFER_APPLICATION_PROTOCOLS"
    unsigned int _SECBUFFER_SRTP_PROTECTION_PROFILES "SECBUFFER_SRTP_PROTECTION_PROFILES"
    unsigned int _SECBUFFER_SRTP_MASTER_KEY_IDENTIFIER "SECBUFFER_SRTP_MASTER_KEY_IDENTIFIER"
    unsigned int _SECBUFFER_TOKEN_BINDING "SECBUFFER_TOKEN_BINDING"
    unsigned int _SECBUFFER_PRESHARED_KEY "SECBUFFER_PRESHARED_KEY"
    unsigned int _SECBUFFER_PRESHARED_KEY_IDENTITY "SECBUFFER_PRESHARED_KEY_IDENTITY"
    unsigned int _SECBUFFER_DTLS_MTU "SECBUFFER_DTLS_MTU"
    unsigned int _SECBUFFER_SEND_GENERIC_TLS_EXTENSION "SECBUFFER_SEND_GENERIC_TLS_EXTENSION"
    unsigned int _SECBUFFER_SUBSCRIBE_GENERIC_TLS_EXTENSION "SECBUFFER_SUBSCRIBE_GENERIC_TLS_EXTENSION"
    unsigned int _SECBUFFER_FLAGS "SECBUFFER_FLAGS"
    unsigned int _SECBUFFER_TRAFFIC_SECRETS "SECBUFFER_TRAFFIC_SECRETS"
    unsigned int _SECBUFFER_CERTIFICATE_REQUEST_CONTEXT "SECBUFFER_CERTIFICATE_REQUEST_CONTEXT"

    unsigned int _SECBUFFER_ATTRMASK "SECBUFFER_ATTRMASK"
    unsigned int _SECBUFFER_READONLY "SECBUFFER_READONLY"
    unsigned int _SECBUFFER_READONLY_WITH_CHECKSUM "SECBUFFER_READONLY_WITH_CHECKSUM"
    unsigned int _SECBUFFER_RESERVED "SECBUFFER_RESERVED"

    cdef struct _SEC_CHANNEL_BINDINGS:
        unsigned int  dwInitiatorAddrType
        unsigned int  cbInitiatorLength
        unsigned int  dwInitiatorOffset
        unsigned int  dwAcceptorAddrType
        unsigned int  cbAcceptorLength
        unsigned int  dwAcceptorOffset
        unsigned int  cbApplicationDataLength
        unsigned int  dwApplicationDataOffset
    ctypedef _SEC_CHANNEL_BINDINGS SEC_CHANNEL_BINDINGS
    ctypedef SEC_CHANNEL_BINDINGS *PSEC_CHANNEL_BINDINGS

SECBUFFER_VERSION = _SECBUFFER_VERSION

class SecBufferFlags(enum.IntFlag):
    SECBUFFER_NONE = 0
    SECBUFFER_ATTRMASK = _SECBUFFER_ATTRMASK
    SECBUFFER_READONLY = _SECBUFFER_READONLY
    SECBUFFER_READONLY_WITH_CHECKSUM = _SECBUFFER_READONLY_WITH_CHECKSUM
    SECBUFFER_RESERVED = _SECBUFFER_RESERVED

class SecBufferType(enum.IntEnum):
    SECBUFFER_EMPTY = _SECBUFFER_EMPTY
    SECBUFFER_DATA = _SECBUFFER_DATA
    SECBUFFER_TOKEN = _SECBUFFER_TOKEN
    SECBUFFER_PKG_PARAMS = _SECBUFFER_PKG_PARAMS
    SECBUFFER_MISSING = _SECBUFFER_MISSING
    SECBUFFER_EXTRA = _SECBUFFER_EXTRA
    SECBUFFER_STREAM_TRAILER = _SECBUFFER_STREAM_TRAILER
    SECBUFFER_STREAM_HEADER = _SECBUFFER_STREAM_HEADER
    SECBUFFER_NEGOTIATION_INFO = _SECBUFFER_NEGOTIATION_INFO
    SECBUFFER_PADDING = _SECBUFFER_PADDING
    SECBUFFER_STREAM = _SECBUFFER_STREAM
    SECBUFFER_MECHLIST = _SECBUFFER_MECHLIST
    SECBUFFER_MECHLIST_SIGNATURE = _SECBUFFER_MECHLIST_SIGNATURE
    SECBUFFER_TARGET = _SECBUFFER_TARGET
    SECBUFFER_CHANNEL_BINDINGS = _SECBUFFER_CHANNEL_BINDINGS
    SECBUFFER_CHANGE_PASS_RESPONSE = _SECBUFFER_CHANGE_PASS_RESPONSE
    SECBUFFER_TARGET_HOST = _SECBUFFER_TARGET_HOST
    SECBUFFER_ALERT = _SECBUFFER_ALERT
    SECBUFFER_APPLICATION_PROTOCOLS = _SECBUFFER_APPLICATION_PROTOCOLS
    SECBUFFER_SRTP_PROTECTION_PROFILES = _SECBUFFER_SRTP_PROTECTION_PROFILES
    SECBUFFER_SRTP_MASTER_KEY_IDENTIFIER = _SECBUFFER_SRTP_MASTER_KEY_IDENTIFIER
    SECBUFFER_TOKEN_BINDING = _SECBUFFER_TOKEN_BINDING
    SECBUFFER_PRESHARED_KEY = _SECBUFFER_PRESHARED_KEY
    SECBUFFER_PRESHARED_KEY_IDENTITY = _SECBUFFER_PRESHARED_KEY_IDENTITY
    SECBUFFER_DTLS_MTU = _SECBUFFER_DTLS_MTU
    SECBUFFER_SEND_GENERIC_TLS_EXTENSION = _SECBUFFER_SEND_GENERIC_TLS_EXTENSION
    SECBUFFER_SUBSCRIBE_GENERIC_TLS_EXTENSION = _SECBUFFER_SUBSCRIBE_GENERIC_TLS_EXTENSION
    SECBUFFER_FLAGS = _SECBUFFER_FLAGS
    SECBUFFER_TRAFFIC_SECRETS = _SECBUFFER_TRAFFIC_SECRETS
    SECBUFFER_CERTIFICATE_REQUEST_CONTEXT = _SECBUFFER_CERTIFICATE_REQUEST_CONTEXT


cdef class SecBufferDesc:
    # cdef _SecBufferDesc raw
    # cdef SecBuffer[:] _buffers

    def __cinit__(
        SecBufferDesc self,
        buffers: list[SecBuffer],
        *,
        unsigned int version = SECBUFFER_VERSION,
    ):
        self._buffers = buffers
        self.raw.ulVersion = version
        self.raw.cBuffers = len(buffers)
        if self.raw.cBuffers:
            self.raw.pBuffers = <PSecBuffer>calloc(sizeof(_SecBuffer), self.raw.cBuffers)
            if not self.raw.pBuffers:
                raise MemoryError("Cannot calloc SecBufferDesc buffers")

            for idx, buffer in enumerate(buffers):
                src = (<SecBuffer>buffer).raw

                self.raw.pBuffers[idx].cbBuffer = src.cbBuffer
                self.raw.pBuffers[idx].BufferType = src.BufferType
                self.raw.pBuffers[idx].pvBuffer = src.pvBuffer

    def __dealloc__(SecBufferDesc self):
        if self.raw.pBuffers:
            free(self.raw.pBuffers)
            self.raw.pBuffers = NULL
            self.raw.cBuffers = 0
        self._buffers = []

    def __iter__(SecBufferDesc self) -> list[SecBuffer]:
        return self._buffers.__iter__()

    def __len__(SecBufferDesc self) -> int:
        return self.raw.cBuffers

    def __getitem__(SecBufferDesc self, key: int) -> SecBuffer:
        return self._buffers[key]

    cdef void mark_as_allocated(SecBufferDesc self):
        for buffer in self._buffers:
            (<SecBuffer>buffer)._needs_free = 1

    cdef void sync_buffers(SecBufferDesc self):
        for idx in range(self.raw.cBuffers):
            (<SecBuffer>self._buffers[idx]).raw.cbBuffer = self.raw.pBuffers[idx].cbBuffer
            (<SecBuffer>self._buffers[idx]).raw.BufferType = self.raw.pBuffers[idx].BufferType
            (<SecBuffer>self._buffers[idx]).raw.pvBuffer = self.raw.pBuffers[idx].pvBuffer

    @property
    def version(SecBuffer self) -> int:
        return self.raw.ulVersion

cdef class SecBuffer:
    # cdef _SecBuffer raw
    # cdef unsigned char[:] _buffer
    # cdef int _needs_free

    def __cinit__(
        SecBuffer self,
        unsigned char[:] data,
        buffer_type: SecBufferType | int,
        buffer_flags: SecBufferFlags | int = 0,
    ):
        self.raw.BufferType = int(buffer_type) | int(buffer_flags)

        if data is not None and len(data):
            self._buffer = data
            self.raw.cbBuffer = <unsigned int>len(data)
            self.raw.pvBuffer = &self._buffer[0]
        else:
            self._buffer = None
            self.raw.cbBuffer = 0
            self.raw.pvBuffer = NULL

    def __dealloc__(SecBuffer self):
        if self.raw.pvBuffer and self._needs_free:
            FreeContextBuffer(self.raw.pvBuffer)
            self._needs_free = 0

        self.raw.pvBuffer = NULL
        self.raw.cbBuffer = 0
        self.raw.BufferType = _SECBUFFER_EMPTY

    def __repr__(SecBuffer self) -> str:
        kwargs = [f"{k}={v}" for k, v in {
            'data': f"bytearray({self.data!r})",
            'buffer_type': self.buffer_type,
            'buffer_flags': self.buffer_flags,
        }.items()]

        return f"SecBuffer({', '.join(kwargs)})"

    def __str__(SecBuffer self) -> str:
        val = self.buffer_type.name
        buffer_flags = self.buffer_flags
        if buffer_flags:
            val = f"{val}|{buffer_flags.name}"

        return val

    @property
    def count(SecBuffer self) -> int:
        return self.raw.cbBuffer

    @property
    def data(SecBuffer self) -> bytes:
        return bytes(self.dangerous_get_view())

    @property
    def buffer_type(SecBuffer self) -> SecBufferType:
        return SecBufferType(self.raw.BufferType & ~_SECBUFFER_ATTRMASK)

    @property
    def buffer_flags(SecBuffer self) -> SecBufferFlags:
        return SecBufferFlags(self.raw.BufferType & _SECBUFFER_ATTRMASK)

    def dangerous_get_view(SecBuffer self) -> memoryview:
        if self.raw.pvBuffer == NULL or self.raw.cbBuffer == 0:
            return memoryview(b"")
        else:
            return memoryview(<char[:self.raw.cbBuffer]>self.raw.pvBuffer)


cdef class SecChannelBindings:
    cdef PSEC_CHANNEL_BINDINGS raw

    def __init__(
        SecChannelBindings self,
        *,
        unsigned int initiator_addr_type = 0,
        const unsigned char[:] initiator_addr = None,
        unsigned int acceptor_addr_type = 0,
        const unsigned char[:] acceptor_addr = None,
        const unsigned char[:] application_data = None,
    ):
        offset = sizeof(SEC_CHANNEL_BINDINGS)

        initiator_offset = 0
        initiator_len = 0
        if initiator_addr is not None:
            initiator_offset = offset
            initiator_len = len(initiator_addr)
            offset += initiator_len

        acceptor_offset = 0
        acceptor_len = 0
        if acceptor_addr is not None:
            acceptor_offset = offset
            acceptor_len = len(acceptor_addr)
            offset += acceptor_len

        application_offset = 0
        application_len = 0
        if application_data is not None:
            application_offset = offset
            application_len = len(application_data)

        raw_length = sizeof(SEC_CHANNEL_BINDINGS) + initiator_len + acceptor_len + application_len
        self.raw = <PSEC_CHANNEL_BINDINGS>malloc(raw_length)
        if not self.raw:
            raise MemoryError("Cannot malloc SecChannelBindings buffers")

        cdef unsigned char[:] raw_ptr = <unsigned char[:raw_length]><unsigned char*>self.raw

        self.raw.dwInitiatorAddrType = initiator_addr_type
        self.raw.cbInitiatorLength = initiator_len
        self.raw.dwInitiatorOffset = initiator_offset
        if initiator_offset:
            raw_ptr[initiator_offset : initiator_offset + initiator_len] = initiator_addr.copy()

        self.raw.dwAcceptorAddrType = acceptor_addr_type
        self.raw.cbAcceptorLength = acceptor_len
        self.raw.dwAcceptorOffset = acceptor_offset
        if acceptor_offset:
            raw_ptr[acceptor_offset : acceptor_offset + acceptor_len] = acceptor_addr.copy()

        self.raw.cbApplicationDataLength = application_len
        self.raw.dwApplicationDataOffset = application_offset
        if application_offset:
            raw_ptr[application_offset : application_offset + application_len] = application_data.copy()

    def __dealloc__(SecChannelBindings self):
        if self.raw:
            free(self.raw)
            self.raw = NULL

    cdef char[:] get_raw_view(SecChannelBindings self):
        data_len = sizeof(SEC_CHANNEL_BINDINGS) + \
            self.raw.cbInitiatorLength + \
            self.raw.cbAcceptorLength + \
            self.raw.cbApplicationDataLength
        return <char[:data_len]><char*>self.raw

    def __repr__(SecChannelBindings self) -> str:
        kwargs = [f"{k}={v}" for k, v in {
            'initiator_addr_type': self.initiator_addr_type,
            'initiator_addr': repr(self.initiator_addr),
            'acceptor_addr_type': self.acceptor_addr_type,
            'acceptor_addr': repr(self.acceptor_addr),
            'application_data': repr(self.application_data),
        }.items()]

        return f"SecChannelBindings({', '.join(kwargs)})"

    @property
    def initiator_addr_type(SecChannelBindings self) -> int:
        return self.raw.dwInitiatorAddrType

    @property
    def initiator_addr(SecChannelBindings self) -> bytes | None:
        val_len = self.raw.cbInitiatorLength
        val_offset = self.raw.dwInitiatorOffset

        if val_offset:
            return (<char *>self.raw)[val_offset : val_offset + val_len]

        else:
            return None

    @property
    def acceptor_addr_type(SecChannelBindings self) -> int | None:
        return self.raw.dwAcceptorAddrType

    @property
    def acceptor_addr(SecChannelBindings self) -> bytes:
        val_len = self.raw.cbAcceptorLength
        val_offset = self.raw.dwAcceptorOffset

        if val_offset:
            return (<char *>self.raw)[val_offset : val_offset + val_len]

        else:
            return None

    @property
    def application_data(SecChannelBindings self) -> bytes | None:
        val_len = self.raw.cbApplicationDataLength
        val_offset = self.raw.dwApplicationDataOffset

        if val_offset:
            return (<char *>self.raw)[val_offset : val_offset + val_len]

        else:
            return None

    def get_sec_buffer_copy(SecChannelBindings self) -> SecBuffer:
        data = bytearray(self.get_raw_view())
        return SecBuffer(data, _SECBUFFER_CHANNEL_BINDINGS)

    def dangerous_get_sec_buffer(SecChannelBindings self) -> SecBuffer:
        data_len = sizeof(SEC_CHANNEL_BINDINGS) + \
            self.raw.cbInitiatorLength + \
            self.raw.cbAcceptorLength + \
            self.raw.cbApplicationDataLength
        view = memoryview(<char[:data_len]><char*>self.raw)

        return SecBuffer(view, _SECBUFFER_CHANNEL_BINDINGS)



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import enum

from ._credential import CredHandle

class KdcProxySettingsFlags(enum.IntEnum):
    """Flags used for :class:`SecPkgCredKdcProxySettings`"""

    KDX_PROXY_SETTINGS_FLAGS_NONE = ...
    """No flags are set."""

    KDC_PROXY_SETTINGS_FLAGS_FORCEPROXY = ...
    """
    Force the use of the proxy specified and do not attempt to talk to the KDC
    directly.
    """

class SecPkgCred:
    """Base class for credential attribute types."""

class SecPkgCredKdcProxySettings(SecPkgCred):
    """Kerberos proxy settings.

    Specifies the kerberos proxy settings for the credentials. The proxy_server
    is in the format of ``hostname:port:path`` where ``port:path`` can be
    omitted for the default of ``443:KdcProxy``.

    This wraps the `SecPkgCredentials_KdcProxySettingsW`_ Win32 structure.

    .. _SecPkgCredentials_KdcProxySettingsW:
        https://learn.microsoft.com/en-us/windows/win32/api/sspi/ns-sspi-secpkgcredentials_kdcproxysettingsw
    """

    def __init__(
        self,
        *,
        flags: int = 0,
        proxy_server: str | None = None,
    ) -> None: ...
    @property
    def version(self) -> int:
        """The version of the internal structure."""
    @property
    def flags(self) -> KdcProxySettingsFlags:
        """Flags for the Kerberos proxy settings."""
    @property
    def proxy_server(self) -> str | None:
        """The proxy server value."""

def set_credentials_attributes(
    context: CredHandle,
    attribute: SecPkgCred,
) -> None:
    """Sets the attributes of a credential.

    Sets the attributes of a credential, such as the name associated with the
    credential. The information is valid for any security context created with
    the specified credential. The credential can be created with
    :meth:`acquire_credentials_handle`.

    The following attributes have been implemented:

        :class:`SecPkgCredKdcProxySettings`

    This wraps the `SetCredentialsAttributesW`_ Win32 function.

    Args:
        credential: The credential to set the attribute on.
        attribute: The attribute to set.

    Raises:
        WindowsError: If the function failed.

    .. _SetCredentialsAttributesW:
        https://learn.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-setcredentialsattributesw
    """



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import enum

from ._security_buffer cimport PSecBufferDesc, SecBufferDesc
from ._security_context cimport CtxtHandle
from ._win32_types cimport *


cdef extern from "python_sspi.h":
    unsigned int _KERB_WRAP_NO_ENCRYPT "KERB_WRAP_NO_ENCRYPT"

    unsigned int _SECQOP_WRAP_NO_ENCRYPT "SECQOP_WRAP_NO_ENCRYPT"
    unsigned int _SECQOP_WRAP_OOB_DATA "SECQOP_WRAP_OOB_DATA"

    # https://learn.microsoft.com/en-us/windows/win32/secauthn/decryptmessage--general
    SECURITY_STATUS DecryptMessage(
        PCtxtHandle    phContext,
        PSecBufferDesc pMessage,
        unsigned int  MessageSeqNo,
        unsigned int  *pfQOP
    ) nogil

    # https://learn.microsoft.com/en-us/windows/win32/secauthn/encryptmessage--general
    SECURITY_STATUS EncryptMessage(
        PCtxtHandle    phContext,
        unsigned int  fQOP,
        PSecBufferDesc pMessage,
        unsigned int  MessageSeqNo
    ) nogil

    # https://learn.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-makesignature
    SECURITY_STATUS MakeSignature(
        PCtxtHandle    phContext,
        unsigned int  fQOP,
        PSecBufferDesc pMessage,
        unsigned int  MessageSeqNo
    ) nogil

    # https://learn.microsoft.com/en-us/windows/win32/api/sspi/nf-sspi-verifysignature
    SECURITY_STATUS VerifySignature(
        PCtxtHandle    phContext,
        PSecBufferDesc pMessage,
        unsigned int  MessageSeqNo,
        unsigned int  *pfQOP
    ) nogil


class QopFlags(enum.IntFlag):
    KERB_WRAP_NO_ENCRYPT = _KERB_WRAP_NO_ENCRYPT
    SECQOP_WRAP_NO_ENCRYPT = _SECQOP_WRAP_NO_ENCRYPT
    SECQOP_WRAP_OOB_DATA = _SECQOP_WRAP_OOB_DATA

def decrypt_message(
    CtxtHandle context not None,
    SecBufferDesc message not None,
    unsigned int seq_no,
) -> int:
    cdef unsigned int qop = 0
    with nogil:
        res = DecryptMessage(
            &context.raw,
            &message.raw,
            seq_no,
            &qop
        )

    if res:
        PyErr_SetFromWindowsErr(res)

    message.sync_buffers()

    return qop

def encrypt_message(
    CtxtHandle context not None,
    unsigned int qop,
    SecBufferDesc message not None,
    unsigned int seq_no,
) -> None:
    with nogil:
        res = EncryptMessage(
            &context.raw,
            qop,
            &message.raw,
            seq_no,
        )

    if res:
        PyErr_SetFromWindowsErr(res)

    message.sync_buffers()

def make_signature(
    CtxtHandle context not None,
    unsigned int qop,
    SecBufferDesc message not None,
    unsigned int seq_no,
) -> None:
    with nogil:
        res = MakeSignature(
            &context.raw,
            qop,
            &message.raw,
            seq_no,
        )

    if res:
        PyErr_SetFromWindowsErr(res)

    message.sync_buffers()

def verify_signature(
    CtxtHandle context not None,
    SecBufferDesc message not None,
    unsigned int seq_no,
) -> int:
    cdef unsigned int qop = 0
    with nogil:
        res = VerifySignature(
            &context.raw,
            &message.raw,
            seq_no,
            &qop
        )

    if res:
        PyErr_SetFromWindowsErr(res)

    message.sync_buffers()

    return qop



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import datetime

EPOCH_FILETIME = 116444736000000000  # 1970-01-01 as FILETIME.


def filetime_to_datetime(
    filetime: int,
) -> datetime.datetime:
    # Expiry is 100 nanosecond intervals since 1601-01-01. It might also
    # exceed the bounds Python can represent which is ignored.
    ft_epoch_ms = (filetime - EPOCH_FILETIME) // 10
    epoch_delta = datetime.timedelta(microseconds=ft_epoch_ms)

    utc = datetime.timezone.utc
    try:
        return datetime.datetime(1970, 1, 1, tzinfo=utc) + epoch_delta
    except OverflowError:
        return datetime.datetime.fromtimestamp(0, tz=utc)



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import datetime
import typing as t

import sspi.raw as raw

from ._filetime import filetime_to_datetime


class UserCredential(raw.CredHandle):
    """A user credential.

    This represents an SSPI credential backed by a username and password. It
    can also be used to specify the security protocol used, defaults to
    ``Negotiate``.

    If using a username in the UPN format ``user@DOMAIN.COM``, do not specify
    anything for the domain argument.

    The usage must be set to ``initiate`` for client credentials, ``accept``
    for server credentials, or ``both`` for both a client and server.

    The protocol_list can be set to a list of protocols that the ``Negotiate``
    protocol can use. To stop a protocol from being used, prefix the value with
    ``!``, for example the following will turn off NTLM being negotiated:

        UserCredential(protocol="Negotiate", protocol_list["!ntlm"])

    This class is designed to be a high level overlay on top of the
    :class:`sspi.raw.CredHandle` class. It can be used as a credential for both
    the high level API as well as te low level API if more complex scenarios
    are needed.

    Args:
        username: The username to authenticate as.
        password: The password for the user specified.
        domain: The domain the user belongs to.
        protocol: The security protocol this credential can use for
            authentication.
        usage: How the credentials will be used.
        protocol_list: A list of protocols to allow or deny.
    """

    def __init__(
        self,
        username: str | None = None,
        password: str | None = None,
        domain: str | None = None,
        protocol: str = "Negotiate",
        usage: t.Literal["initiate", "accept", "both"] = "initiate",
        *,
        protocol_list: list[str] | None = None,
    ) -> None:
        cred_use = {
            "initiate": raw.CredentialUse.SECPKG_CRED_OUTBOUND,
            "accept": raw.CredentialUse.SECPKG_CRED_INBOUND,
            "both": raw.CredentialUse.SECPKG_CRED_BOTH,
        }[usage]
        temp_cred, expiry = raw.acquire_credentials_handle(
            principal=None,
            package=protocol,
            credential_use=cred_use,
            auth_data=raw.WinNTAuthIdentity(
                username=username,
                domain=domain,
                password=password,
                package_list=":".join(protocol_list) if protocol_list else None,
            ),
        )
        # Transfer ownership of the acquired cred to this instance.
        raw._credential._replace_cred_handle(temp_cred, self)

        self._expiry = filetime_to_datetime(expiry)
        self._protocol = protocol
        self._username = username or "CurrentUser"
        if domain:
            self._username = f"{domain}\\{self._username}"

    def __str__(self) -> str:
        return f"{self.username} - {self.protocol}"

    @property
    def expiry(self) -> datetime.datetime:
        """The time when this credential expires."""
        return self._expiry

    @property
    def protocol(self) -> str:
        """The security protocol this credential can use."""
        return self._protocol

    @property
    def username(self) -> str:
        """The username this credential is for."""
        return self._username



#!/usr/bin/env python

# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import ctypes
import ctypes.util
import os
import pathlib
import platform
import typing

from Cython.Build import cythonize
from setuptools import Extension, setup

SKIP_EXTENSIONS = os.environ.get("SSPI_SKIP_EXTENSIONS", "false").lower() == "true"
SKIP_MODULE_CHECK = os.environ.get("SSPI_SKIP_MODULE_CHECK", "false").lower() == "true"
CYTHON_LINETRACE = os.environ.get("SSPI_CYTHON_TRACING", "false").lower() == "true"
SSPI_MAIN_LIB = os.environ.get("SSPI_MAIN_LIB", None)


def make_extension(
    name: str,
    module: ctypes.CDLL | None,
    canary: typing.Optional[str] = None,
    **kwargs: typing.Any,
) -> Extension | None:
    source = pathlib.Path("src") / (name.replace(".", os.sep) + ".pyx")

    if not SKIP_MODULE_CHECK and canary and not hasattr(module, canary):
        print(f"Skipping {source} as it is not supported by the selected SSPI implementation.")
        return None

    if not source.exists():
        raise FileNotFoundError(source)

    print(f"Compiling {source}")
    return Extension(
        name=name,
        sources=[str(source)],
        **kwargs,
    )


raw_extensions = []
if not SKIP_EXTENSIONS:
    if os.name == "nt":
        windir = pathlib.Path(os.environ.get("SystemRoot", r"C:\Windows"))
        sspi_path = str((windir / "System32" / "Secur32.dll").absolute())

        define_macros = [
            ("UNICODE", "1"),
            ("_UNICODE", "1"),
            ("SECURITY_WIN32", "1"),
        ]
        extra_compile_args = []
        sspi_lib = "Secur32"
        text_libs = []

    else:
        ext = "dylib" if platform.system() == "Darwin" else "so"
        sspi_path = f"libsspi.{ext}"

        define_macros = []
        extra_compile_args = ["-DPYSSPI_IS_LINUX"]
        sspi_lib = "sspi"
        text_libs = ["icuuc"]

    if SSPI_MAIN_LIB:
        sspi_path = SSPI_MAIN_LIB

    if SKIP_MODULE_CHECK:
        sspi = None
    else:
        print(f"Using {sspi_path} as SSPI module for platform checks")
        sspi = ctypes.CDLL(sspi_path)

    for e in [
        "context_attributes",
        "credential_attributes",
        "credential",
        "message",
        "ntstatus",
        "security_buffer",
        "security_context",
        "security_package",
        ("text", text_libs),
    ]:
        name = e
        libraries = [sspi_lib]
        canary = None
        if isinstance(e, tuple):
            name = e[0]
            if len(e) > 1:
                libraries.extend(e[1])
            if len(e) > 2:
                canary = e[2]

        ext = make_extension(
            f"sspi.raw._{name}",
            module=sspi,
            canary=canary,
            extra_compile_args=extra_compile_args,
            libraries=libraries,
            define_macros=define_macros,
        )
        if ext:
            raw_extensions.append(ext)

setup(
    ext_modules=cythonize(
        raw_extensions,
        language_level=3,
        compiler_directives={"linetrace": CYTHON_LINETRACE},
    ),
)



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from ._win32_types cimport *


cdef class CredHandle:
    cdef _CredHandle raw
    cdef int needs_free



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

ERROR_MAP = {
    0x00090312: (
        "SEC_I_CONTINUE_NEEDED",
        "The function completed successfully, but it must be called again to complete the context.",
    ),
    0x00090313: ("SEC_I_COMPLETE_NEEDED", "The function completed successfully, but CompleteToken must be called."),
    0x00090314: (
        "SEC_I_COMPLETE_AND_CONTINUE",
        "The function completed successfully, but both CompleteToken and this function must be called to complete the context.",
    ),
    0x00090315: (
        "SEC_I_LOCAL_LOGON",
        "The logon was completed, but no network authority was available. The logon was made using locally known information.",
    ),
    0x00090317: ("SEC_I_CONTEXT_EXPIRED", "The context has expired and can no longer be used."),
    0x00090320: (
        "SEC_I_INCOMPLETE_CREDENTIALS",
        "The credentials supplied were not complete and could not be verified. Additional information can be returned from the context.",
    ),
    0x00090321: ("SEC_I_RENEGOTIATE", "The context data must be renegotiated with the peer."),
    0x00090323: ("SEC_I_NO_LSA_CONTEXT", "There is no LSA mode context associated with this context."),
    0x0009035C: (
        "SEC_I_SIGNATURE_NEEDED",
        "A signature operation must be performed before the user can authenticate.",
    ),
    0x80090300: ("SEC_E_INSUFFICIENT_MEMORY", "Not enough memory is available to complete this request."),
    0x80090301: ("SEC_E_INVALID_HANDLE", "The handle specified is invalid."),
    0x80090302: ("SEC_E_UNSUPPORTED_FUNCTION", "The function requested is not supported."),
    0x80090303: ("SEC_E_TARGET_UNKNOWN", "The specified target is unknown or unreachable."),
    0x80090304: ("SEC_E_INTERNAL_ERROR", "The Local Security Authority (LSA) cannot be contacted."),
    0x80090305: ("SEC_E_SECPKG_NOT_FOUND", "The requested security package does not exist."),
    0x80090306: ("SEC_E_NOT_OWNER", "The caller is not the owner of the desired credentials."),
    0x80090307: ("SEC_E_CANNOT_INSTALL", "The security package failed to initialize and cannot be installed."),
    0x80090308: ("SEC_E_INVALID_TOKEN", "The token supplied to the function is invalid."),
    0x80090309: (
        "SEC_E_CANNOT_PACK",
        "The security package is not able to marshal the logon buffer, so the logon attempt has failed.",
    ),
    0x8009030A: (
        "SEC_E_QOP_NOT_SUPPORTED",
        "The per-message quality of protection is not supported by the security package.",
    ),
    0x8009030B: ("SEC_E_NO_IMPERSONATION", "The security context does not allow impersonation of the client."),
    0x8009030C: ("SEC_E_LOGON_DENIED", "The logon attempt failed."),
    0x8009030D: ("SEC_E_UNKNOWN_CREDENTIALS", "The credentials supplied to the package were not recognized."),
    0x8009030E: ("SEC_E_NO_CREDENTIALS", "No credentials are available in the security package."),
    0x8009030F: ("SEC_E_MESSAGE_ALTERED", "The message or signature supplied for verification has been altered."),
    0x80090310: ("SEC_E_OUT_OF_SEQUENCE", "The message supplied for verification is out of sequence."),
    0x80090311: ("SEC_E_NO_AUTHENTICATING_AUTHORITY", "No authority could be contacted for authentication."),
    0x80090316: ("SEC_E_BAD_PKGID", "The requested security package does not exist."),
    0x80090317: ("SEC_E_CONTEXT_EXPIRED", "The context has expired and can no longer be used."),
    0x80090318: ("SEC_E_INCOMPLETE_MESSAGE", "The supplied message is incomplete. The signature was not verified."),
    0x80090320: (
        "SEC_E_INCOMPLETE_CREDENTIALS",
        "The credentials supplied were not complete and could not be verified. The context could not be initialized.",
    ),
    0x80090321: ("SEC_E_BUFFER_TOO_SMALL", "The buffers supplied to a function were too small."),
    0x80090322: ("SEC_E_WRONG_PRINCIPAL", "The target principal name is incorrect."),
    0x80090324: ("SEC_E_TIME_SKEW", "The clocks on the client and server machines are skewed."),
    0x80090325: ("SEC_E_UNTRUSTED_ROOT", "The certificate chain was issued by an authority that is not trusted."),
    0x80090326: ("SEC_E_ILLEGAL_MESSAGE", "The message received was unexpected or badly formatted."),
    0x80090327: ("SEC_E_CERT_UNKNOWN", "An unknown error occurred while processing the certificate."),
    0x80090328: ("SEC_E_CERT_EXPIRED", "The received certificate has expired."),
    0x80090329: ("SEC_E_ENCRYPT_FAILURE", "The specified data could not be encrypted."),
    0x80090330: ("SEC_E_DECRYPT_FAILURE", "The specified data could not be decrypted."),
    0x80090331: (
        "SEC_E_ALGORITHM_MISMATCH",
        "The client and server cannot communicate because they do not possess a common algorithm.",
    ),
    0x80090332: (
        "SEC_E_SECURITY_QOS_FAILED",
        "The security context could not be established due to a failure in the requested quality of service (for example, mutual authentication or delegation).",
    ),
    0x80090333: (
        "SEC_E_UNFINISHED_CONTEXT_DELETED",
        "A security context was deleted before the context was completed. This is considered a logon failure.",
    ),
    0x80090334: (
        "SEC_E_NO_TGT_REPLY",
        "The client is trying to negotiate a context and the server requires user-to-user but did not send a ticket granting ticket (TGT) reply.",
    ),
    0x80090335: (
        "SEC_E_NO_IP_ADDRESSES",
        "Unable to accomplish the requested task because the local machine does not have IP addresses.",
    ),
    0x80090336: (
        "SEC_E_WRONG_CREDENTIAL_HANDLE",
        "The supplied credential handle does not match the credential associated with the security context.",
    ),
    0x80090337: (
        "SEC_E_CRYPTO_SYSTEM_INVALID",
        "The cryptographic system or checksum function is invalid because a required function is unavailable.",
    ),
    0x80090338: ("SEC_E_MAX_REFERRALS_EXCEEDED", "The number of maximum ticket referrals has been exceeded."),
    0x80090339: (
        "SEC_E_MUST_BE_KDC",
        "The local machine must be a Kerberos domain controller (KDC), and it is not.",
    ),
    0x8009033A: (
        "SEC_E_STRONG_CRYPTO_NOT_SUPPORTED",
        "The other end of the security negotiation requires strong cryptography, but it is not supported on the local machine.",
    ),
    0x8009033B: ("SEC_E_TOO_MANY_PRINCIPALS", "The KDC reply contained more than one principal name."),
    0x8009033C: (
        "SEC_E_NO_PA_DATA",
        "Expected to find PA data for a hint of what etype to use, but it was not found.",
    ),
    0x8009033D: (
        "SEC_E_PKINIT_NAME_MISMATCH",
        "The client certificate does not contain a valid user principal name (UPN), or does not match the client name in the logon request. Contact your administrator.",
    ),
    0x8009033E: ("SEC_E_SMARTCARD_LOGON_REQUIRED", "Smart card logon is required and was not used."),
    0x8009033F: ("SEC_E_SHUTDOWN_IN_PROGRESS", "A system shutdown is in progress."),
    0x80090340: ("SEC_E_KDC_INVALID_REQUEST", "An invalid request was sent to the KDC."),
    0x80090341: (
        "SEC_E_KDC_UNABLE_TO_REFER",
        "The KDC was unable to generate a referral for the service requested.",
    ),
    0x80090342: ("SEC_E_KDC_UNKNOWN_ETYPE", "The encryption type requested is not supported by the KDC."),
    0x80090343: (
        "SEC_E_UNSUPPORTED_PREAUTH",
        "An unsupported pre-authentication mechanism was presented to the Kerberos package.",
    ),
    0x80090345: (
        "SEC_E_DELEGATION_REQUIRED",
        "The requested operation cannot be completed. The computer must be trusted for delegation, and the current user account must be configured to allow delegation.",
    ),
    0x80090346: (
        "SEC_E_BAD_BINDINGS",
        "Client's supplied Security Support Provider Interface (SSPI) channel bindings were incorrect.",
    ),
    0x80090347: ("SEC_E_MULTIPLE_ACCOUNTS", "The received certificate was mapped to multiple accounts."),
    0x80090348: ("SEC_E_NO_KERB_KEY", "No Kerberos key was found."),
    0x80090349: ("SEC_E_CERT_WRONG_USAGE", "The certificate is not valid for the requested usage."),
    0x80090350: (
        "SEC_E_DOWNGRADE_DETECTED",
        "The system detected a possible attempt to compromise security. Ensure that you can contact the server that authenticated you.",
    ),
    0x80090351: (
        "SEC_E_SMARTCARD_CERT_REVOKED",
        "The smart card certificate used for authentication has been revoked. Contact your system administrator. The event log might contain additional information.",
    ),
    0x80090352: (
        "SEC_E_ISSUING_CA_UNTRUSTED",
        "An untrusted certification authority (CA) was detected while processing the smart card certificate used for authentication. Contact your system administrator.",
    ),
    0x80090353: (
        "SEC_E_REVOCATION_OFFLINE_C",
        "The revocation status of the smart card certificate used for authentication could not be determined. Contact your system administrator.",
    ),
    0x80090354: (
        "SEC_E_PKINIT_CLIENT_FAILURE",
        "The smart card certificate used for authentication was not trusted. Contact your system administrator.",
    ),
    0x80090355: (
        "SEC_E_SMARTCARD_CERT_EXPIRED",
        "The smart card certificate used for authentication has expired. Contact your system administrator.",
    ),
    0x80090356: (
        "SEC_E_NO_S4U_PROT_SUPPORT",
        "The Kerberos subsystem encountered an error. A service for user protocol requests was made against a domain controller that does not support services for users.",
    ),
    0x80090357: (
        "SEC_E_CROSSREALM_DELEGATION_FAILURE",
        "An attempt was made by this server to make a Kerberos-constrained delegation request for a target outside the server's realm. This is not supported and indicates a misconfiguration on this server's allowed-to-delegate-to list. Contact your administrator.",
    ),
    0x80090358: (
        "SEC_E_REVOCATION_OFFLINE_KDC",
        "The revocation status of the domain controller certificate used for smart card authentication could not be determined. The system event log contains additional information. Contact your system administrator.",
    ),
    0x80090359: (
        "SEC_E_ISSUING_CA_UNTRUSTED_KDC",
        "An untrusted CA was detected while processing the domain controller certificate used for authentication. The system event log contains additional information. Contact your system administrator.",
    ),
    0x8009035A: (
        "SEC_E_KDC_CERT_EXPIRED",
        "The domain controller certificate used for smart card logon has expired. Contact your system administrator with the contents of your system event log.",
    ),
    0x8009035B: (
        "SEC_E_KDC_CERT_REVOKED",
        "The domain controller certificate used for smart card logon has been revoked. Contact your system administrator with the contents of your system event log.",
    ),
    0x8009035D: ("SEC_E_INVALID_PARAMETER", "One or more of the parameters passed to the function were invalid."),
    0x8009035E: (
        "SEC_E_DELEGATION_POLICY",
        "The client policy does not allow credential delegation to the target server.",
    ),
    0x8009035F: (
        "SEC_E_POLICY_NLTM_ONLY",
        "The client policy does not allow credential delegation to the target server with NLTM only authentication.",
    ),
}


class WindowsError(OSError):
    """WindowsError

    This is the Exception class that replicates WindowsError that is present
    only on Windows. It manually maps known error codes to helpful error
    messages. It is only used outside of Windows where it isn't already
    defined.

    Args:
        winerror: The SSPI HRESULT error code to raise.
    """

    def __init__(self, winerror: int) -> None:
        hresult = ((1 << 32) + winerror) if winerror < 0 else winerror
        status_id, status_msg = ERROR_MAP.get(hresult, ("UNKNOWN_ERROR_CODE", "Error is unknown"))

        strerror = f"Received SSPI Error {status_id} 0x{hresult:08X}: {status_msg}"
        super().__init__(winerror, strerror)
        self.winerror = winerror



# Copyright: (c) 2023, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import enum


class NtStatus(enum.IntEnum):
    SEC_I_COMPLETE_AND_CONTINUE = _SEC_I_COMPLETE_AND_CONTINUE
    SEC_I_COMPLETE_NEEDED = _SEC_I_COMPLETE_NEEDED
    SEC_I_CONTINUE_NEEDED = _SEC_I_CONTINUE_NEEDED
    SEC_E_OK = _SEC_E_OK

    @classmethod
    def _missing_(cls, value: object) -> typing.Optional[enum.Enum]:
        if not isinstance(value, int):
            return None

        new_member = int.__new__(cls)
        new_member._name_ = "Unknown NtStatus Value 0x{0:08X}".format(value)
        new_member._value_ = value
        return cls._value2member_map_.setdefault(value, new_member)



# Copyright: (c) 2023 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import sys

from ._context_attributes import (
    SecPkgContext,
    SecPkgContextNames,
    SecPkgContextPackageInfo,
    SecPkgContextSessionKey,
    SecPkgContextSizes,
    query_context_attributes,
)
from ._credential import (
    AcquireCredentialsResult,
    AuthIdentity,
    CredentialUse,
    CredHandle,
    WinNTAuthFlags,
    WinNTAuthIdentity,
    acquire_credentials_handle,
)
from ._credential_attributes import (
    KdcProxySettingsFlags,
    SecPkgCred,
    SecPkgCredKdcProxySettings,
    set_credentials_attributes,
)
from ._message import (
    QopFlags,
    decrypt_message,
    encrypt_message,
    make_signature,
    verify_signature,
)
from ._ntstatus import NtStatus
from ._security_buffer import (
    SECBUFFER_VERSION,
    SecBuffer,
    SecBufferDesc,
    SecBufferFlags,
    SecBufferType,
    SecChannelBindings,
)
from ._security_context import (
    AcceptContextResult,
    AscReq,
    AscRet,
    CtxtHandle,
    InitializeContextResult,
    IscReq,
    IscRet,
    TargetDataRep,
    accept_security_context,
    complete_auth_token,
    initialize_security_context,
)
from ._security_package import (
    SecPkgInfo,
    SecurityPackageCapability,
    enumerate_security_packages,
)

if sys.platform == "win32":
    WindowsError = WindowsError
else:
    from ._winerror import WindowsError

__all__ = [
    "SECBUFFER_VERSION",
    "AcceptContextResult",
    "AscReq",
    "AscRet",
    "AcquireCredentialsResult",
    "AuthIdentity",
    "CredentialUse",
    "CredHandle",
    "CtxtHandle",
    "InitializeContextResult",
    "IscReq",
    "IscRet",
    "KdcProxySettingsFlags",
    "NtStatus",
    "QopFlags",
    "SecBuffer",
    "SecBufferDesc",
    "SecBufferFlags",
    "SecBufferType",
    "SecChannelBindings",
    "SecPkgContext",
    "SecPkgContextNames",
    "SecPkgContextPackageInfo",
    "SecPkgContextSessionKey",
    "SecPkgContextSizes",
    "SecPkgCred",
    "SecPkgCredKdcProxySettings",
    "SecPkgInfo",
    "SecurityPackageCapability",
    "TargetDataRep",
    "WindowsError",
    "WinNTAuthFlags",
    "WinNTAuthIdentity",
    "accept_security_context",
    "acquire_credentials_handle",
    "complete_auth_token",
    "decrypt_message",
    "encrypt_message",
    "enumerate_security_packages",
    "initialize_security_context",
    "make_signature",
    "query_context_attributes",
    "set_credentials_attributes",
    "verify_signature",
]
